﻿using System.Text;
using System.Text.Json;

namespace HDLNoCGen
{
    class GraphMesh : Graph, IRectangleLike
    {
        public override string getSignature()
        {
            return "Mesh";
        }

        protected List<List<int>> ways_routing_XY { get; set; }

        protected double efficiencyXY { get; set; }

        protected double average_distance_XY { get;  set; }

        public GraphMesh() : base()
        {
            this.ways_routing_XY = null;
            this.graphId = GraphType.Mesh;
        }

        public GraphMesh(GraphDeserializer graph_des) : base(graph_des)
        {

        }

        public GraphMesh(string paraneters, CancellationToken token)
        {
            this.graphId = GraphType.Mesh;
            this.node_count = 1;

            string[] buffer = paraneters.Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
            string[] buffer_topology = buffer[0].Trim().Split(new string[] { "M", "m", ";", ":", ", ", ")", " " }, StringSplitOptions.RemoveEmptyEntries);
            generators = new List<int>();

            for (int i = 0; i < buffer_topology.Length && i < 2; i++)
            {
                generators.Add(Convert.ToInt32(buffer_topology[i]));
                this.node_count *= generators[i];
            }

            if (generators.Min() < 2)
            {
                throw new Exception("Неправильно задана топология");
            }


            generators.Sort();

            diameter = generators[0];

            string[] buffer_parameters;
            if (buffer.Length > 1)
            {
                buffer_parameters = buffer[1].Trim().Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries);
                buffer_parameters[0].Trim();
                string[] sub_buffer_parameters = buffer_parameters[0].Trim().Split(new string[] { "=" }, StringSplitOptions.RemoveEmptyEntries);
                //this.p = Convert.ToInt32(sub_buffer_parameters[1]);
                this.node_count = Convert.ToInt32(buffer_parameters[1]) * Convert.ToInt32(buffer_parameters[2]);
            }

            //string[] buffer = paraneters.Split(new string[] { "С(", "C(", ";", ":",  ", ", ")", " " }, StringSplitOptions.RemoveEmptyEntries);
            //this.node_count = Convert.ToInt32(buffer[0]);
            //this.p = Convert.ToInt32(param_2);



            this.is_created = true;
        }

        protected override void SetNodeCountFromGenerators(List<int> generators)
        {
            this.node_count = generators[0] * generators[1];
        }

        public List<List<int>> Get_ways_routing_XY()
        {
            return this.ways_routing_XY;
        }

        public double getXY_Efficiency(CancellationToken token)
        {
            calculateEfficiencyXY(token);
            return this.efficiencyXY;
        }

        private void calculateEfficiencyXY (CancellationToken token)
        {
            double deikstra_count = 0;
            double XY_count = 0;

            for (int i = 0; i < this.min_ways_routing_deikstra.Count; i++)
            {
                if (token.IsCancellationRequested)
                {
                    token.ThrowIfCancellationRequested();
                }
                deikstra_count += this.min_ways_routing_deikstra[i].Count - 1;
                XY_count += this.ways_routing_XY[i].Count - 1;
            }

            this.efficiencyXY = Math.Round(deikstra_count / XY_count, 3);
        }

        public List<List<int>> Generate_XY_routing(CancellationToken token)
        {
            int iterations_count; // для того чтобы программа не зависала, если не может посчитать маршрут

            int start_node;
            int end_node;
            List<int> route;

            this.ways_routing_XY = new List<List<int>>();

            try
            {
                if (!this.is_created)
                {
                    throw new Exception("Описание топологии не было задано или задано не полностью");
                }

                for (int i = 0; i < this.node_count; i++)
                {
                    for (int j = 0; j < this.node_count; j++)
                    {
                        if (token.IsCancellationRequested)
                        {
                            token.ThrowIfCancellationRequested();
                        }
                        iterations_count = 0;
                        start_node = i;
                        end_node = j;
                        route = new List<int>();
                        route.Add(start_node + Settings.Get_node_naming_start_index());
                        while (start_node != end_node)
                        {
                            iterations_count++;

                            start_node = XY_routing(start_node, end_node, generators, node_count);
                            route.Add(start_node + Settings.Get_node_naming_start_index());

                            if (iterations_count > Settings.Get_error_iterations_count())
                            {
                                throw new Exception("Превышено количество итераций моделирования маршрута для алгоритма Simple");
                            }
                        }
                        this.ways_routing_XY.Add(route);
                    }
                }
            }
            catch (Exception ex)
            {
                this.error_message = ex.Message;
                return null;
            }

            return this.ways_routing_XY;
        } // проверка на ошибки добавлена

        private int XY_routing(int start_node, int end_node, List<int> generators, int node_count)
        {
            int S = end_node - start_node;
            if (S == 0)
            {
                return (start_node);
            }
            else if ((start_node % generators[1]) < (end_node % generators[1]))
            {
                start_node++;
            }
            else if ((start_node % generators[1]) > (end_node % generators[1]))
            {
                start_node--;
            }
            else if ((((start_node % generators[0]) < (end_node % generators[0])) && (start_node % node_count) < (end_node % node_count)) || ((start_node % node_count) < (end_node % node_count)))
            {
                start_node += generators[1];
            }
            else if ((start_node % generators[0]) > (end_node % generators[0]) || ((start_node % node_count) > (end_node % node_count)))
            {
                start_node -= generators[1];
            }
            //if (start_node == 0) 
            //start_node = this.node_count;
            return (start_node);
        }

        public double Calculate_average_distance_XY(CancellationToken token)
        {
            average_distance_XY = 0;

            for (int i = 0; i < this.ways_routing_XY.Count; i++)
            {
                if (token.IsCancellationRequested)
                {
                    token.ThrowIfCancellationRequested();
                }
                average_distance_XY += ways_routing_XY[i].Count;
            }
            average_distance_XY /= ways_routing_XY.Count;

            return average_distance_XY;
        }

        protected override void Create_matr_smej()
        {
            this.matr_smej = new int[this.node_count, this.node_count];
            int n = generators[0];
            int m = generators[1];

            for (int i = 0; i < this.node_count; i++)
            {
                if (i % m != 0)
                {
                    matr_smej[i, i - 1] = 1;
                    matr_smej[i - 1, i] = 1;
                }
                if (i / m != 0)
                {
                    matr_smej[i, i - m] = 1;
                    matr_smej[i - m, i] = 1;
                }
            }
        }

        protected override int GetCordinatesSize()
        {
            int coordinates_size = 0;
            int counter = 1;
            while (counter < generators.Max())
            {
                coordinates_size++;
                counter *= 2;
            }
            return coordinates_size;
        }

        protected override string descipher_router(RouterOptions.Algorithm obj)
        {
            string response = "..\\..\\..\\verilog-primitives\\src\\alg\\mesh\\";

            switch (obj)
            {
                case RouterOptions.Algorithm.XY:
                    response += "XY.sv";
                    break;
                default:
                    throw new NotImplementedException($"Algorithm {obj} was not implemented into this switch statement.");
            }

            return response;
        }

        public override void createRouter(string project_path, router_options parametrs)
        {
            base.createRouter(project_path, parametrs);

            string router_format = descipher_router(parametrs.queue_position);
            string router_text = "";
            string router_json = File.ReadAllText("..\\..\\..\\router_variants.json");
            var router_variants = JsonSerializer.Deserialize<Dictionary<string, Dictionary<string, Dictionary<string, List<string>>>>>(router_json);

            switch (parametrs.algorithm, parametrs.queue_position)
            {
                case (RouterOptions.Algorithm.XY, RouterOptions.Queue_position.Front):
                    router_text = String.Format(router_format, String.Join("\n    ", router_variants["mesh"]["xy"]["front"].Select(s => $"{s}")).Split("c#_splitter"));
                    break;
                case (RouterOptions.Algorithm.XY, RouterOptions.Queue_position.Rear):
                    router_text = String.Format(router_format, String.Join("\n    ", router_variants["mesh"]["xy"]["rear"].Select(s => $"{s}")).Split("c#_splitter"));
                    break;
                default:
                    throw new NotImplementedException($"Router format {parametrs.algorithm}, {parametrs.queue_position} was not implemented into this switch statement.");
            }

            File.WriteAllText(project_path + "\\src\\router.sv", router_text);


            string file_to_format = "noc";
            string file_text = "";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format+".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text, 
                parametrs.packet_width,
                GetCordinatesSize(), 
                node_count
                ));

            file_to_format = "noc_XY";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                generators[0], generators[1]
                ));

            file_to_format = "queue";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                parametrs.queue_length, get_bits(parametrs.queue_length)
                ));

            file_to_format = "router";
            file_text = File.ReadAllText("..\\..\\..\\verilog-primitives\\inc\\" + file_to_format + ".svh");
            File.WriteAllText(project_path + "\\inc\\" + file_to_format + ".svh", String.Format(file_text,
                5,
                this.get_bits(5)
                ));


        }

        public override void createNoC(string project_path, router_options parameters)
        {
            base.createNoC(project_path, parameters);

            File.Copy("..\\..\\..\\verilog-primitives\\noc\\mesh.sv", project_path + "\\noc\\noc.sv");
            File.Copy("..\\..\\..\\verilog-primitives\\sel\\rectangle.sv", project_path + "\\noc\\toplevel.sv");
            File.Copy("..\\..\\..\\verilog-primitives\\tb\\routing\\rectangle.sv", project_path + "\\tb\\tb_routing.sv");

            switch (parameters.queue_position)
            {
                case RouterOptions.Queue_position.Front:
                    File.Copy("..\\..\\..\\verilog-primitives\\tb\\queueing\\front\\rectangle.sv", project_path + "\\tb\\tb_queueing.sv");
                    break;
                case RouterOptions.Queue_position.Rear:
                    File.Copy("..\\..\\..\\verilog-primitives\\tb\\queueing\\rear\\rectangle.sv", project_path + "\\tb\\tb_queueing.sv");
                    break;
            }
        }
    }
}

