﻿using Npgsql;
using NpgsqlTypes;
using System.Diagnostics;
using System.Globalization;
using System.Management;
using System.Reflection.Metadata;
using System.Text;

namespace HDLNoCGen
{

    public struct ParsedReport
    {
        public string graphType { get; set; }
        public string algorithm { get; set; }
        public int nodeCount { get; set; }
        public string generators { get; set; }
        public int infoLength { get; set; }
        public double efficiency { get; set; }
        public double averageDistance { get; set; }
        public int numberOfALMs { get; set; }
        public int numberOfRegisters { get; set; }

    }

    static class ProjectGenerator
    {
        #region legacy code

        public static void initializeTCL(string project_name, string routing_algorithm, string graph_signature, List<int> generators)
        {
            string top_entity_name = $"top_level_{project_name}";
            string router_name = $"router_{project_name}";
            string data_selector_name = $"data_selector_{project_name}";

            FileStream fs;
            StreamWriter sw;

            Directory.CreateDirectory($"{Settings.Get_project_path()}{project_name}");
            fs = new FileStream($"{Settings.Get_project_path()}{project_name}\\Setup.tcl", FileMode.Create, FileAccess.ReadWrite);
            sw = new StreamWriter(fs);

            sw.WriteLine($"project_new {project_name} -overwrite");
            sw.WriteLine();
            sw.WriteLine("set_global_assignment -name FAMILY \"Cyclone V\"");
            sw.WriteLine("set_global_assignment -name DEVICE 5CGXFC9E7F35C8");
            sw.WriteLine($"set_global_assignment -name VERILOG_FILE {data_selector_name}.v");
            sw.WriteLine($"set_global_assignment -name VERILOG_FILE {top_entity_name}.v");
            sw.WriteLine($"set_global_assignment -name VERILOG_FILE {router_name}.v");
            sw.WriteLine($"set_global_assignment -name TOP_LEVEL_ENTITY {top_entity_name}");
            sw.WriteLine();
            sw.WriteLine("load_package flow");
            sw.WriteLine("execute_flow -compile");
            sw.WriteLine();
            sw.WriteLine("project_close");

            sw.Close();
            fs.Close();
        }

        public static void createBat(string project_name)
        {
            FileStream fs;
            StreamWriter sw;

            Directory.CreateDirectory($"{Settings.Get_project_path()}{project_name}");
            fs = new FileStream($"{Settings.Get_project_path()}{project_name}\\Compile.bat", FileMode.Create, FileAccess.ReadWrite);
            sw = new StreamWriter(fs);

            sw.WriteLine("C:");
            sw.WriteLine("cd \\");
            sw.WriteLine($"cd {Settings.Get_project_path()}{project_name}");
            sw.WriteLine($"{Settings.Get_path_to_quartus()}quartus_sh -t Setup.tcl");

            sw.Close();
            fs.Close();

        }
        #endregion

        public static void killProcessAndChildren(int pid)
        {
            if (pid == 0)
            {
                return;
            }
            ManagementObjectSearcher searcher = new ManagementObjectSearcher("Select * From Win32_Process Where ParentProcessID=" + pid);
            ManagementObjectCollection moc = searcher.Get();
            foreach (ManagementObject mo in moc)
            {
                killProcessAndChildren(Convert.ToInt32(mo["ProcessID"]));
            }
            try
            {
                Process proc = Process.GetProcessById(pid);
                proc.Kill();
            }
            catch (ArgumentException)
            {
                // Process already exited.
            }
            catch (System.ComponentModel.Win32Exception)
            {
                //Access rejected
            }

            GC.Collect();

        }

        /// <summary>
        /// Функция для генерации tcl
        /// </summary>
        /// <param name="project_path">Путь к корневой папке проекта.</param>
        /// <param name="project_name">Название проекта.</param>
        /// <param name="toplevel_entity_name">Название top level entity.</param>
        public static void createSetupSH(string project_path, string project_name, string toplevel_entity_name)
        {
            var fs = new FileStream($"{project_path}\\Setup_sh.tcl", FileMode.Create, FileAccess.ReadWrite);
            var sw = new StreamWriter(fs);

            string top_entity_folder = toplevel_entity_name switch
            {
                "noc" => "noc",
                "toplevel" => "noc",
                "router" => "src",
                _ => throw new Exception($"Toplevel '{toplevel_entity_name}' does not exist. It should be either 'noc' or 'router'.")
            };

            sw.WriteLine($"project_new {project_name} -overwrite");
            sw.WriteLine();
            sw.WriteLine("set_global_assignment -name FAMILY \"Cyclone V\"");
            sw.WriteLine("set_global_assignment -name DEVICE 5CGXFC9E7F35C8");
            sw.WriteLine($"set_global_assignment -name TOP_LEVEL_ENTITY {toplevel_entity_name}");
            sw.WriteLine($"set_global_assignment -name SYSTEMVERILOG_FILE {top_entity_folder}/{toplevel_entity_name}.sv");
            sw.WriteLine();
            sw.WriteLine("load_package flow");
            sw.WriteLine("execute_flow -compile");
            sw.WriteLine();
            sw.WriteLine("project_close");

            sw.Close();
            fs.Close();
        }

        /// <summary>
        /// Функция для генерации tcl
        /// </summary>
        /// <param name="project_path">Путь к корневой папке проекта.</param>
        /// <param name="project_name">Название проекта.</param>
        public static void createSetupSTA(string project_path, string project_name)
        {
            var fs = new FileStream($"{project_path}\\Setup_sta.tcl", FileMode.Create, FileAccess.ReadWrite);
            var sw = new StreamWriter(fs);

            sw.WriteLine($"project_open {project_name}");
            sw.WriteLine();
            sw.WriteLine("create_timing_netlist -model slow ");
            sw.WriteLine("read_sdc");
            sw.WriteLine("update_timing_netlist");
            sw.WriteLine();
            sw.WriteLine("set outfile1 [open \"clocker.logger\" w+] ");
            sw.WriteLine("  puts $outfile1 [get_clock_fmax_info]");
            sw.WriteLine("close $outfile1");
            sw.WriteLine();
            sw.WriteLine("delete_timing_netlist");
            sw.WriteLine();
            sw.WriteLine("project_close");

            sw.Close();
            fs.Close();
        }

        public static void createSimulateScript(string project_path, string project_name, string mode)
        {
            var fs = new FileStream($"{project_path}\\Modelsim_simulate.tcl", FileMode.Create, FileAccess.ReadWrite);
            var sw = new StreamWriter(fs);

            sw.WriteLine($"vlib work");
            sw.WriteLine($"vlog -sv tb/tb_{mode}.sv");
            sw.WriteLine($"vsim work.tb_{mode}");
            sw.WriteLine($"run -all");

            sw.Close();
            fs.Close();
        }

        public static Tuple<string, Process, Process> routerPreparation(Graph graph, string project_name, string project_path, router_options parametres)
        {
            string path = $"{project_path}\\{project_name}";

            graph.createRouter(path, parametres);
            FileStream fs;
            StreamWriter sw;

            fs = new FileStream($"{path}\\Compile.bat", FileMode.Create, FileAccess.ReadWrite);
            sw = new StreamWriter(fs);

            sw.WriteLine($"{path.Substring(0, 2)}");
            sw.WriteLine($"cd {path}\n");
            sw.WriteLine($"{Settings.Get_path_to_quartus()}quartus_sh -t Setup_sh.tcl");
            sw.WriteLine($"{Settings.Get_path_to_quartus()}quartus_sta -t Setup_sta.tcl");

            sw.Close();
            fs.Close();


            fs = new FileStream($"{path}\\Simulate.bat", FileMode.Create, FileAccess.ReadWrite);
            sw = new StreamWriter(fs);

            sw.WriteLine($"{path.Substring(0, 2)}");
            sw.WriteLine($"cd {path}\n");
            sw.WriteLine($"vsim -c -do Modelsim_simulate.tcl");

            sw.Close();
            fs.Close();

            Process simulation_process = new Process();
            Process compilation_process = new Process();

            Console.CancelKeyPress += delegate {
                killProcessAndChildren(compilation_process.Id);
                killProcessAndChildren(simulation_process.Id);
            };

            string comp_path = path + "\\Compile.bat";
            string sim_path = path + "\\Simulate.bat";

            compilation_process.StartInfo.FileName = comp_path;
            compilation_process.StartInfo.UseShellExecute = false;
            compilation_process.StartInfo.RedirectStandardOutput = true;
            compilation_process.StartInfo.CreateNoWindow = true;
            compilation_process.EnableRaisingEvents = true;
            compilation_process.StartInfo.StandardOutputEncoding = Encoding.GetEncoding("utf-8");

            simulation_process.StartInfo.FileName = sim_path;
            simulation_process.StartInfo.UseShellExecute = false;
            simulation_process.StartInfo.RedirectStandardOutput = true;
            simulation_process.StartInfo.CreateNoWindow = true;
            simulation_process.EnableRaisingEvents = true;
            simulation_process.StartInfo.StandardOutputEncoding = Encoding.GetEncoding("utf-8");

            return new Tuple<string, Process, Process>(path, compilation_process, simulation_process);
        }

        public static void networkCompilation(CancellationToken GeneralPurposeCancellationToken,
            Process compile, Process simulate,
            Graph graph, router_options parameters,
            string path, string project_name,
            bool create_database_entry, bool cleanup)
        {

            List<int> generators = graph.generators;
            string graph_signature = "";
            switch (graph.graphId)
            {
                case GraphType.Circulant:
                    graph_signature = "c";
                    break;
                case GraphType.Mesh:
                    graph_signature = "m";
                    break;
                case GraphType.Torus:
                    graph_signature = "t";
                    break;
            };


            string simulation_report = "";
            string[] separators = { "Errors: ", ", Warnings: " };

            DataReceivedEventHandler simulationTextReader = (s, ev) =>
            {
                if (!String.IsNullOrEmpty(ev.Data))
                {
                    simulation_report += $"{ev.Data}\n";
                    Console.WriteLine($"{ev.Data}\n");
                }
            };

            EventHandler simulationFinishingHandler = (s, ev) =>
            {
                List<string> sliced_report = simulation_report.Split(separators, System.StringSplitOptions.None).ToList();

                if (!sliced_report[1].Equals("0") || !sliced_report[3].Equals("0"))
                {
                    FileStream fs = new FileStream($"{Settings.Get_project_path()}\\simulation_error_logs.log", FileMode.Append, FileAccess.Write);
                    StreamWriter sw = new StreamWriter(fs);

                    sw.WriteLine($"{graph_signature} {String.Join(" ", generators)}, Errors: {sliced_report[0]}, Warnings: {sliced_report[1]}");

                    sw.Close();
                    fs.Close();

                    killProcessAndChildren(simulate.Id);

                    throw new Exception("Simulation error");
                }

            };

            simulate.OutputDataReceived += simulationTextReader;
            simulate.Exited += simulationFinishingHandler;

            createSimulateScript(path, project_name, "routing");

            simulate.Start();
            simulate.BeginOutputReadLine();

            simulate.WaitForExit();
            simulate.CancelOutputRead();


            createSimulateScript(path, project_name, "queueing");

            simulate.Start();
            simulate.BeginOutputReadLine();

            simulate.WaitForExit();
            simulate.CancelOutputRead();


            double algorithmEf = 0;
            double algorithmAd = 0;
            EventHandler compilationFinishingHandler;

            switch (graph.graphId)
            {
                case GraphType.Circulant:

                    GraphCirculant graphCirculant = graph as GraphCirculant;

                    graphCirculant.Generate_GA_routing(GeneralPurposeCancellationToken);
                    algorithmEf = graphCirculant.Calculate_efficiency_GA(GeneralPurposeCancellationToken);
                    algorithmAd = graphCirculant.Calculate_average_distance_GA(GeneralPurposeCancellationToken);

                    break;
                case GraphType.Mesh:

                    GraphMesh graphMesh = graph as GraphMesh;

                    graphMesh.Generate_XY_routing(GeneralPurposeCancellationToken);
                    algorithmEf = graphMesh.getXY_Efficiency(GeneralPurposeCancellationToken);
                    algorithmAd = graphMesh.Calculate_average_distance_XY(GeneralPurposeCancellationToken);
                    break;
                case GraphType.Torus:

                    GraphTorus graphTorus = graph as GraphTorus;

                    graphTorus.Generate_XY_routing(GeneralPurposeCancellationToken);
                    algorithmEf = graphTorus.getXY_Efficiency(GeneralPurposeCancellationToken);
                    algorithmAd = graphTorus.Calculate_average_distance_XY(GeneralPurposeCancellationToken);
                    break;
                default:
                    throw new NotImplementedException($"Graph type {graph.graphId} was not implemented into this switch statement.");
            }

            //Подготовлен запуск
            createSetupSH(path, project_name, "toplevel");
            createSetupSTA(path, project_name);

            if (create_database_entry)
            {
                string queue_type = parameters.queue_type switch
                {
                    RouterOptions.Queue_types.Index => "index",
                    RouterOptions.Queue_types.Pointer => "pointer",
                    RouterOptions.Queue_types.Line => "line",
                    _ => "-"
                };
                string queue_position = parameters.queue_position switch
                {
                    RouterOptions.Queue_position.Front => "front",
                    RouterOptions.Queue_position.Rear => "rear",
                    _ => "-"
                };
                string arb_type = parameters.arbiter_type switch
                {
                    RouterOptions.Arbiter_types.Round_Robin => "round_robin",
                    _ => "-"
                };
                string algorithm = parameters.algorithm switch
                {
                    RouterOptions.Algorithm.XY => "xy",
                    RouterOptions.Algorithm.GA => "ga",
                    _ => "-"
                };

                string signature = graph.getSignature();
                ParsedReport parsedReport = new ParsedReport
                {
                    graphType = signature,
                    algorithm = algorithm,
                    nodeCount = graph.node_count,
                    generators = String.Join(", ", generators.Select(s => $"{s}")),
                    infoLength = Convert.ToInt32(parameters.info_width),
                    efficiency = algorithmEf,
                    averageDistance = algorithmAd,
                    numberOfALMs = -1,
                    numberOfRegisters = -1
                };

                List<string> fitmentSummary = new List<string>(0);
                compilationFinishingHandler = (s, ev) =>
                {
                    FileStream fs = new FileStream($"{path}\\{project_name}.fit.summary", FileMode.Open, FileAccess.Read);
                    StreamReader sr = new StreamReader(fs);

                    string line;
                    while ((line = sr.ReadLine()) != null)
                    {
                        fitmentSummary.Add($"{line}");
                    }

                    sr.Close();
                    fs.Close();


                    int alm_start_index = fitmentSummary[7].IndexOf(':') + 2;
                    int alm_length = fitmentSummary[7].IndexOf('/') - 1 - alm_start_index;
                    List<string> separated_number = new List<string>(fitmentSummary[7].Substring(alm_start_index, alm_length).Split(','));
                    int thousand_separator = 1;

                    for (int i = separated_number.Count - 1; i >= 0; i--)
                    {
                        parsedReport.numberOfALMs += Int32.Parse(separated_number[i]) * thousand_separator;
                        thousand_separator *= 1000;
                    }

                    int reg_start_index = fitmentSummary[8].IndexOf(':') + 2;
                    parsedReport.numberOfRegisters = Int32.Parse(fitmentSummary[8].Substring(reg_start_index));

                    //mark
                    var connectionString = $"Server={Settings.server};Port={Settings.port};Database={Settings.database};User Id={Settings.username};Password={Settings.password}";

                    var connection = new NpgsqlConnection(connectionString);
                    connection.Open();

                    var createTableCommand = new NpgsqlCommand("CREATE TABLE IF NOT EXISTS TestingReportsNetwork (" +
                         "graph_type VARCHAR(20)," +
                         "buffers VARCHAR(20)," +
                         "buffers_position VARCHAR(20)," +
                         "buffers_number INTEGER," +
                         "arbiter VARCHAR(20)," +
                         "algorithm VARCHAR(20)," +
                         "generators VARCHAR(20)," +
                         "info_length INTEGER," +
                         "ALMs INTEGER," +
                         "Registers INTEGER," +
                         "Frequency REAL," +
                         "PRIMARY KEY(graph_type, buffers, buffers_position, buffers_number, arbiter, algorithm, generators, info_length)" +
                         ")", connection);
                    createTableCommand.ExecuteNonQuery();

                    var insertCommand = new NpgsqlCommand("INSERT INTO TestingReportsNetwork (graph_type, buffers, buffers_position, buffers_number, arbiter, algorithm, generators, info_length, ALMs, Registers, Frequency)" +
                        "VALUES (@graph_type, @buffers, @buffers_position, @buffers_number, @arbiter, @algorithm, @generators, @info_length, @ALMs, @Registers, @Frequency)", connection);
                    insertCommand.Parameters.Add(new NpgsqlParameter("graph_type", NpgsqlDbType.Varchar));
                    insertCommand.Parameters.Add(new NpgsqlParameter("buffers", NpgsqlDbType.Varchar));
                    insertCommand.Parameters.Add(new NpgsqlParameter("buffers_position", NpgsqlDbType.Varchar));
                    insertCommand.Parameters.Add(new NpgsqlParameter("buffers_number", NpgsqlDbType.Integer));
                    insertCommand.Parameters.Add(new NpgsqlParameter("arbiter", NpgsqlDbType.Varchar));
                    insertCommand.Parameters.Add(new NpgsqlParameter("algorithm", NpgsqlDbType.Varchar));
                    insertCommand.Parameters.Add(new NpgsqlParameter("generators", NpgsqlDbType.Varchar));
                    insertCommand.Parameters.Add(new NpgsqlParameter("info_length", NpgsqlDbType.Integer));
                    insertCommand.Parameters.Add(new NpgsqlParameter("ALMs", NpgsqlDbType.Integer));
                    insertCommand.Parameters.Add(new NpgsqlParameter("Registers", NpgsqlDbType.Integer));
                    insertCommand.Parameters.Add(new NpgsqlParameter("Frequency", NpgsqlDbType.Real));


                    insertCommand.Parameters["graph_type"].Value = parsedReport.graphType;
                    insertCommand.Parameters["algorithm"].Value = parsedReport.algorithm;
                    insertCommand.Parameters["generators"].Value = parsedReport.generators;
                    insertCommand.Parameters["info_length"].Value = parameters.info_width;
                    insertCommand.Parameters["ALMs"].Value = parsedReport.numberOfALMs;
                    insertCommand.Parameters["Registers"].Value = parsedReport.numberOfRegisters;

                    insertCommand.Parameters["buffers"].Value = queue_type;
                    insertCommand.Parameters["buffers_position"].Value = queue_position;
                    insertCommand.Parameters["buffers_number"].Value = Convert.ToInt32(parameters.queue_length);
                    insertCommand.Parameters["arbiter"].Value = arb_type;

                    fs = new FileStream($"{path}\\clocker.logger", FileMode.Open, FileAccess.Read);
                    sr = new StreamReader(fs);

                    insertCommand.Parameters["Frequency"].Value = float.Parse(sr.ReadLine().Split(' ')[1], new CultureInfo("en-US"));

                    sr.Close();
                    fs.Close();

                    insertCommand.ExecuteNonQuery();

                    connection.Close();

                    Console.WriteLine("Запись в базу данных прошла успешно\n");
                    //mark

                };
            }
            else compilationFinishingHandler = (s, ev) => { };

            compile.Exited += compilationFinishingHandler;

            compile.Start();
            compile.BeginOutputReadLine();

            compile.WaitForExit();
            Console.WriteLine("Process exited with code " + compile.ExitCode);
        }
    }
}

